#!/bin/sh
# $Id: udev.get_unique_hardware_path.sh,v 1.3 2004/06/08 11:50:56 harald Exp $
#set -x

# this script should provide the shortest possible unique hardware path 
# to a block device
#

# to be called from udev
# DEVPATH=/block/sda/sda3 $0  (or similar)
# $0 /block/sda
# $0 /sys/block/sda

# example for all:
# for i in `find /sys/block -name dev` ;do DEVPATH="`echo $i | sed -e 's@^/sys\|/dev@@g'`" $0 ; done

# examples:
# SCSI cdrom
# /block/sr0 -> /devices/pci0002:30/0002:30:0c.0/host0/0:0:1:0
# result: pci-0002:30:0c.0-scsi-0:0:1:0
# SCSI disk
# /block/sda -> /devices/pci0002:30/0002:30:0c.0/host0/0:0:4:0
# result: pci-0002:30:0c.0-scsi-0:0:4:0 
# SATA disk, 4 channels per controller
# /block/sda -> /devices/pci0001:00/0001:00:07.0/0001:05:0c.0/host0/0:0:0:0
# result: pci-0001:05:0c.0-scsi-0:0:0:0
# IDE disk
# /block/hda -> /devices/pci0002:02/0002:02:0d.0/ide0/0.0
# result: pci-0002:02:0d.0-ide-0.0
# IDE cdrom on a Mac ASIC:
# /block/hdc -> /devices/pci0001:01/0001:01:17.0/0.80000000:mac-io/0.00020000:ata-3/ide1/1.0
# result: mac-io_ata-3_master
# IDE cdrom on a Mac ASIC, with ide-scsi:
# /block/sr0 -> /devices/pci0001:01/0001:01:17.0/0.80000000:mac-io/0.0001f000:ata-4/ide0/0.1/host2/2:0:0:0
# result: mac-io_ata-4_slave

# USB CDrom drive without 'serial' number:
# reusing 'product' and 'manufacturer' string, if available
# /block/sr0 -> /devices/pci0001:00/0001:00:04.0/0001:02:0b.0/usb4/4-2/4-2:1.0/host4/4:0:0:0
# result: usb-storage-odd-Freecom-USIDERev930:0:0:0

# devices may have several interfaces on one PCI device, like IDE:
# pci-0001:00:04.0_ide1-master
# pci-0001:00:04.0_ide2-master
# pci-0001:00:04.0_ide2-slave
# they are marked as ports, it is expected that the driver shows
# ide1 even if there is nothing connected to either master or slave
# interface
#
# match order is important.
# first IDE to find ide-scsi devices.
# then SCSI
# first usb-storage
# then firewire sbp2
# then the rest

#PATH=/lib/klibc/bin #/usr:/usr/sbin:/usr/bin:/sbin:/bin
SYSFS=/sys
export PATH
RESULT=1
CDROM=
TYPE=

pwd() {
        if [ -x /bin/pwd ]; then
                /bin/pwd;
        else
                pwd;
        fi
}


OPWD="`pwd`"
full_sysfs_class_path=
full_sysfs_device_path=

if [ -z "$DEVPATH" -a -z "$1" ] ; then
    exit 1
fi

if [ -z "$DEVPATH" ] ; then
    case "$1" in
	$SYSFS/*)
	    DEVPATH="${1#$SYSFS}"
	    ;;
	*)
	    DEVPATH=$1
	    ;;
    esac
fi

if [ ! -d $SYSFS$DEVPATH ] ; then
    exit 1
fi
if [ ! -f $SYSFS$DEVPATH/dev ] ; then
    exit 1
fi

case "$DEVPATH" in
    /block/*)
	TYPE=block
	;;
    /class/*)
	TYPE="${DEVPATH#/class/}"
	TYPE="${TYPE%%/*}"
	;;
    *)
	exit 1
	;;
esac
	
#
##
#

get_port () {
    local type offset port
    type=$1
    offset=$2
    for i in $type[0-9]* ; do
	: i $i
	port="${i#$type}"
	if [ "$port" -lt "$offset" ] ; then offset=$port ; fi
    done
    if [ "$port" != "0" ] ; then
	echo $(($2 - $offset))
    fi
}

get_scsi_type () {
    local type

    read type < $1/type
    case "$type" in
	0)
	    # Direct-access device
	    devtype=disk
	    ;;
	1)
	    # Sequential access device
	    devtype=tape
	    ;;
	2)
	    # Printer device
	    devtype=printer
	    ;;
	3)
	    # Processor device
	    devtype=scanner
	    ;;
	4)
	    # Write-once device
	    devtype=cdroms
	    ;;
	5)
	    # CD-ROM device
	    devtype=cdroms
	    ;;
	6)
	    # Scanner device
	    devtype=scanner
	    ;;
	7)
            # Optical memory device
	    devtype=disk
	    ;;
	8)
	    # Medium changer devices
	    devtype=generic
	    ;;
	9)
	    # Communications device
	    devtype=generic
	    ;;
	10|11)
	    # Graphic arts pre-press device
	    devtype=generic
	    ;;
	12)
	    # Storage array controller
	    devtype=generic
	    ;;
	13)
	    # Enclosure service device
	    devtype=generic
	    ;;
	14)
	    # Simplified direct-access device
	    devtype=generic
	    ;;
	15)
	    # Optical card reader/writer device
	    devtype=generic;
	    ;;
	*)
	    devtype=generic;
	    ;;
    esac

    echo $devtype
}

handle_block () {
full_sysfs_class_path="$SYSFS$DEVPATH"
if [ ! -f $full_sysfs_class_path/dev ] ; then return ; fi
# the main device has (hopefully) a symlink to the real device
# a partition is a subdir of the main (raw) device
if [ ! -L $full_sysfs_class_path/device ] ; then
	if [ -f $full_sysfs_class_path/range ] ; then return ; fi
	full_sysfs_class_path="${full_sysfs_class_path%/*}"
	: full_sysfs_class_path "$full_sysfs_class_path"
	if [ ! -L $full_sysfs_class_path/device -o ! -f $full_sysfs_class_path/dev ] ; then
		return
	fi

fi
cd $full_sysfs_class_path/device
full_sysfs_device_path="`pwd`"
cd "$OPWD"
D=$full_sysfs_device_path

devtype=disk
case "$D" in
	*/ide[0-9]/[0-9].[0-9]*|*/ide[0-9][0-9]/[0-9][0-9].[0-9]*)
	# IDE
	: D $D
	# remove ide-scsi part, leave only channel info
	d="`echo $D | sed -e 's@/host[0-9]\{1,\}/[0-9].*@@;s@/ide[0-9]\{1,\}/[0-9]\{1,\}\.@/@'`"
	# port info if the controller has more than one interface
	port="`echo $D | sed -e 's@^.*/ide\([0-9]\{1,\}\).*@\1@'`"
	: d $d kernel_port $port
	case "$d" in
		 */0)
		 channel=0
		 ;;
		 */1)
		 channel=1
		 ;;
		 *)
		 echo "Error: $d is neither master or slave" >&2
	esac
	case "$d" in
		*:mac-io/*)
		: mac-io: $d
		d="`echo $d | sed -e 's@^.*:mac-io[^:]\+:\([^/]\+\).*@mac-io_\1@'`"
		;;
		*)
		d="`echo $d | sed -e 's@^.*/\([^/]\{1,\}\)/.@pci-\1@'`"
		;;
	esac
	if [ -f "$D"/type ] ;then
	        devtype="`get_scsi_type $D`"
	else
		if [ -f /proc/ide/${full_sysfs_class_path##*/}/media ]; then
			read media < /proc/ide/${full_sysfs_class_path##*/}/media
			case "$media" in
			        disk)
				devtype=disk
				;;
				cdrom)
				devtype=cdroms
				;;
			        tape)
				devtype=tape
				;;
			        floppy)
				devtype=disk
				;;
			        *)
				devtype=generic
				;;
			esac
		fi
	fi
	: D $D
	case "$D" in
		*/host[0-9]*)
		D="${D%/*}"
		D="${D%/*}"
		;;
	esac
	cd $D/../..
	port="`get_port ide $port`"
	cd "$OPWD"
	:  hardware_port $port
	if [ -z "$port" ] ; then
		d="${d}-ide-0:$channel"
	else
		d="${d}-ide-${port}:$channel"
	fi
	if [ "$devtype" = "generic" ]; then
	    : not handled
	    RESULT=1
	    return
	else
	    RESULT=0
	fi
	;;
	*/usb[0-9]*/[0-9]*/host[0-9]*/[0-9]*:[0-9]*:[0-9]*:[0-9]*)
	# usb-storage devs have a serial number, hopefully unique
	d="`echo $D |  sed -e 's@/[^/]\{1,\}/host[0-9]\{1,\}/[0-9]\{1,\}:[0-9]\{1,\}:[0-9]\{1,\}:[0-9]\{1,\}@@'`"
	cil="`echo $D | sed -e 's@^.*/[^:]\+@@'`"
	serial=
	if [ -f $D/../../../serial ] ; then
		serial="`sed -e 's@^[ -]\{1,\}\|[ -]\{1,\}$@@g;s@[^abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-_0123456789]@@g' < $D/../../../serial`"
		: serial XXX_${serial}_XXX
		devtype=disk
		d="usb-$serial"
		serial="`echo $serial | sed -e 's@[ 0]\{1,\}@@g'`"
	fi
	if [ -z "$serial" ] ; then
		# no serial, broken device
		# has eventually binary junk in vpd
		identifier=
		if [ -f $D/../../../product ] ; then
		product="`sed -e 's@^[ -]\{1,\}\|[ -]\{1,\}$@@g;s@[^abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-_0123456789]@@g' < $D/../../../product`"
		fi
		if [ -f $D/../../../manufacturer ] ; then
		manufacturer="`sed -e 's@^[ -]\{1,\}\|[ -]\{1,\}$@@g;s@[^abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-_0123456789]@@g' < $D/../../../manufacturer`"
		fi
		if [ -z "$product" -o -z "$manufacturer" ] ; then
			read idvendor < $D/../../../idVendor
			read idproduct < $D/../../../idProduct
			identifier="0x${idvendor}-0x${idproduct}"
		else
			identifier="${manufacturer}-${product}"
		fi
		devtype=disk
		d="usb-${identifier}"
	fi
	d="$d$cil"
	RESULT=0

	;;
	*/css0/*)
	# S/390 zfcp adapter
	if [ -r $full_sysfs_device_path/wwpn ]; then
	    read wwpn < $full_sysfs_device_path/wwpn
	fi
	if [ -r $full_sysfs_device_path/fcp_lun ]; then
	    read lun < $full_sysfs_device_path/fcp_lun
	fi
	if [ -r $full_sysfs_device_path/hba_id ]; then
	    read bus_id < $full_sysfs_device_path/hba_id
	fi
	if [ "$bus_id" -a "$wwpn" -a "$lun" ]; then
	    d="ccw-$bus_id-zfcp-$wwpn:$lun"
	    devtype=disk
	    RESULT=0
	else
	    RESULT=1
	fi
	;;
	*/host[0-9]*/[0-9]*:[0-9]*:[0-9]*:[0-9]*)
	# check for ieee1394 sbp2 
	if test -f $D/ieee1394_id ; then
		read ieee1394_id < $D/ieee1394_id
		d="`echo ieee1394-${ieee1394_id} | sed -e 's@:@-@g'`"
		devtype=disk
		RESULT=0
	else
	# SCSI device
	d="`echo $D |  sed -e 's@/host[0-9]\{1,\}/[0-9]\{1,\}:@/@'`"
	# port on the PCI device
	port="`echo $D |  sed -e 's@.*/host\([0-9]\{1,\}\)/.*@\1@'`"
	cil="`echo $d |  sed -e 's@^.*/@@'`"
	case "$d" in
		# grand central, old powermacs
		*:gc/*)
		adapter="`echo $d |  sed -e 's@/[^/]\{1,\}$@@;s@^.*/@@;s@^.*:@@'`"
		bus="gc"
		;;
		*)
		adapter="`echo $d |  sed -e 's@/[^/]\{1,\}$@@;s@^.*/@@'`"
		bus="pci"
		;;
	esac
	: d $d kernel_port $port
	if [ -d $D/type ]; then
	    devtype="`get_scsi_type $D`"
	fi
	: devtype $devtype
	cd $D/../..
	port="`get_port host $port`"
	cd "$OPWD"
	:  hardware_port $port
	d="$bus-$adapter"
	if [ -z "$port" ] ; then
		d="${d}-scsi-0:${cil}"
	else
		d="${d}-scsi-${port}:${cil}"
	fi
	RESULT=0
	fi

	;;

	*/css0/*)
	# DASD devices
	bus="ccw"
	adapter=${D##*/}
	d="$bus-$adapter"
	;;
	*)
	: not handled
	RESULT=1
	return

	;;
esac
# look for a partition
if [ "$full_sysfs_class_path" != "$SYSFS$DEVPATH" ] ; then
	dp="`echo $SYSFS$DEVPATH | sed -e 's@^/.*/@@;s@^[^0-9]\{1,\}@@;s@.*_@@'`"
	case "$d" in
		*[0-9])
		d="${d}p${dp}"
		;;
		*)
		d="${d}${dp}"
		;;
	esac
fi
# the /dev/cdrom symlink was a bright idea
# it is created during install
# if /dev is a tmpfs mount, print it, otherwise leave it alone
# more than one cdrom device will break because the last device wins
if [ "$devtype" = "cdroms" ] ; then
	C=
	while read a b c ; do
		case "$b" in
			/dev)
			C=$CDROM
			break
			;;
			*)
			;;
		esac
	done < /proc/self/mounts
	CDROM=$C
fi
# done
echo "$devtype/by-path/$d" "$CDROM"
}

case "$TYPE" in
	block)
	handle_block
	;;
	scsi_generic)
	d="`handle_block`"
	RESULT=$?
	echo "$d" | sed -e 's@ .*$@-generic@g'
	;;
	*)
	RESULT=1
	;;
esac
exit $RESULT
